"""
metadata = {
    type: add_argument | add_parser
    args: arguments
    kwargs: keyword arguments
}
"""

import inspect
import logging
from argparse import ArgumentParser, Namespace
from dataclasses import Field, dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Sequence, Tuple, Type, TypeVar

_logger = logging.getLogger(__name__)

T = TypeVar('T', bound='TypedArgs')


def _get_dataclass_fields(cls) -> Dict[str, Field]:
    return getattr(cls, '__dataclass_fields__')


@dataclass
class TypedArgs:

    @classmethod
    def from_args(
        cls: Type[T],
        args: Optional[List[str]] = None,
        namespace: Optional[Namespace] = None,
        parser: Optional[ArgumentParser] = None,
    ) -> T:

        if parser is None:
            parser = ArgumentParser()

        cls._add_arguments(parser, prefix='')

        args = parser.parse_args(args=args, namespace=namespace)

        typed_args = cls()
        typed_args._assign(args, prefix='')

        return typed_args

    @classmethod
    def from_known_args(
        cls: Type[T],
        args: Optional[List[str]] = None,
        namespace: Optional[Namespace] = None,
        parser: Optional[ArgumentParser] = None,
    ) -> Tuple[T, List[str]]:
        if parser is None:
            parser = ArgumentParser()

        cls._add_arguments(parser, prefix='')

        args, unknown = parser.parse_known_args(args=args, namespace=namespace)

        typed_args = cls()
        typed_args._assign(args, prefix='')

        return typed_args, unknown

    @classmethod
    def _add_arguments(
        cls,
        parser: ArgumentParser,
        prefix: str = '',
    ):
        fields = _get_dataclass_fields(cls)
        for name, field in fields.items():

            # 嵌套
            dest = prefix + name
            _logger.debug('enter dest: %s', dest)

            if inspect.isclass(field.type) and issubclass(field.type, TypedArgs):
                # 如果是嵌套的，就加上prefix之后解析
                _logger.debug('add recursive parser: %s', dest)
                field.type._add_arguments(parser, prefix=dest + '.')
            else:
                func_type = field.metadata.get('type', 'none')
                _logger.debug('get metadata=%s', field.metadata)
                if func_type == 'add_argument':
                    args: Sequence[str] = field.metadata['args']
                    kwargs = field.metadata['kwargs']
                    if len(args) > 0 and args[0].startswith('-'):
                        new_args = []
                        for a in args:
                            if a.startswith('--'):
                                new_args.append('--' + prefix + a[2:])
                            else:
                                new_args.append('-' + prefix + a[1:])
                        args = new_args

                    _logger.debug('add normal argument: %s', dest)
                    parser.add_argument(*args, dest=dest, **kwargs)
                elif func_type == 'none':
                    continue
                else:
                    raise Exception('no such type: {}'.format(func_type))

    def _assign(self, args: Namespace, prefix: str = ''):
        fields = _get_dataclass_fields(self)
        for name, field in fields.items():
            dest = prefix + name

            if inspect.isclass(field.type) and issubclass(field.type, TypedArgs):
                if field.default is not None:
                    field_args = field.default
                else:
                    field_args = field.type()
                field_args._assign(args, prefix=name + '.')
                setattr(self, name, field_args)
            else:
                func_type = field.metadata.get('type', 'none')

                if func_type == 'add_argument':
                    setattr(self, name, getattr(args, dest))
                elif func_type == 'none':
                    continue
                else:
                    raise Exception('no such type: {}'.format(func_type))


def add_argument(
    *flags: str,
    action=None,
    nargs=None,
    const=None,
    default=None,
    type=None,
    choices=None,
    required=None,
    help=None,
    metavar=None,
) -> Field:
    kwargs = locals()
    args = kwargs.pop('flags')
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    return _add_argument(*args, **kwargs)


def _add_argument(*args, **kwargs) -> Field:
    """
    Args:
        action: str
    """
    metadata = {'type': 'add_argument', 'args': args, 'kwargs': kwargs}
    default = kwargs.get('default', None)
    _logger.debug('metadata: %s', metadata)
    return field(default=default, metadata=metadata)


def add_parser():
    pass


def _add_parser(**kwargs):
    pass


def _get_members(cls) -> Dict[str, Enum]:
    return getattr(cls, '__members__')


class SubTypedArgs(TypedArgs, Enum):

    @classmethod
    def _add_arguments(cls, parser: ArgumentParser, prefix: str):
        subparsers = parser.add_subparsers()
        members = _get_members(cls)

        for name, member in members.items():
            name = prefix + name
            subparser = subparsers.add_parser(name)
            member.value._add_arguments(subparser)



import sys

if sys.version_info.major == 3 and sys.version_info.minor <= 6:
    import warnings

    warnings.warn(
        'Please `pip install dataclasses` if you are using python 3.6 and below'
    )

from ._typed_args import TypedArgs, add_argument

__version__ = "0.5.0.a2"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typed_args']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'typed-args',
    'version': '0.5.0a2',
    'description': 'Parse command line arguments by defining dataclasses',
    'long_description': '# TypedArgs\n\n[![Build Status](https://travis-ci.org/SunDoge/typed-args.svg?branch=master)](https://travis-ci.org/SunDoge/typed-args)\n[![Version](https://img.shields.io/pypi/v/typed-args)](https://pypi.org/project/typed-args/)\n\nStrong type args.\n\nThis project is inspired by [TeXitoi/structopt](https://github.com/TeXitoi/structopt).\n\n## Install\n\nFrom pypi\n\n```bash\npip install typed-args\n```\n\nIf you want to use it on python 3.5 and 3.6 please install `dataclasses`:\n\n```bash\npip install dataclasses\n```\n\n## Usage\n\n\n```python\nimport argparse\nfrom dataclasses import dataclass\n\nimport typed_args as ta\n\n"""\nargparse\n"""\nparser = argparse.ArgumentParser()\nparser.add_argument(\n    \'data\', metavar=\'DIR\', type=str,\n    help=\'path to dataset\'\n)\nparser.add_argument(\n    \'-a\', \'--arch\', metavar=\'ARCH\', default=\'resnet18\', type=str,\n    help=\'model architecture (default: resnet18)\'\n)\nparser.add_argument(\n    \'-j\', \'--workers\', default=4, metavar=\'N\', type=int, dest=\'num_workers\',\n    help=\'number of data loading workers (default: 4)\'\n)\n\n"""\nTypedArgs\n"""\n\n\n@dataclass\nclass Args(ta.TypedArgs):\n    data: str = ta.add_argument(\n        metavar=\'DIR\', type=str, help=\'path to dataset\'\n    )\n    arch: str = ta.add_argument(\n        \'-a\', \'--arch\', metavar=\'ARCH\', default=\'resnet18\', type=str,\n        help=\'model architecture (default: resnet18)\'\n    )\n    num_workers: int = ta.add_argument(\n        \'-j\', \'--workers\', default=4, metavar=\'N\', type=int,\n        help=\'number of data loading workers (default: 4)\'\n    )\n\n\ndef test_args():\n    data = \'/path/to/dataset\'\n    arch = \'resnet50\'\n    num_workers = 8\n\n    argv = f\'{data} -a {arch} --workers {num_workers}\'.split()\n\n    """\n    from_args = parse_args, from_known_args = parse_known_args\n    """\n    typed_args = Args.from_args(argv)\n    args = parser.parse_args(argv)\n\n    assert args.arch == typed_args.arch\n    assert args.data == typed_args.data\n    assert args.num_workers == typed_args.num_workers\n\n\nif __name__ == "__main__":\n    test_args()\n```\n\n## Limitation\n\nCurrently, we don\'t support `add_group` and `sub parser`, but we will.',
    'author': 'SunDoge',
    'author_email': '384813529@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
